module Ex02 where

import Data.List (sort)
import Data.Semigroup
import Test.QuickCheck



-- PART 1 --

-- In this part you will establish that any semigroup can be
-- "promoted" to a monoid simply by adjoining a new element,
-- and decreeing this new element to be the monoid identity.

-- The elements of `Promote s` consist of the following:
-- 1. all elements of s (in the form `Emb x`)
-- 2. a new element `New`, distinct from all elements of s.

data Promote s
  = Emb s
  | New
  deriving (Show, Eq)

instance (Arbitrary s) => Arbitrary (Promote s) where
  arbitrary =
    (\f x -> f x) <$> oneof [return (const New), return Emb]
                  <*> arbitrary

-- Task 1.1: Define a monoid operation `promoteOp` on
-- the type `Promote g` so that the adjoined element `New`
-- functions as the identity element of the resulting monoid.

promoteOp :: (Semigroup g) => Promote g -> Promote g -> Promote g
promoteOp (Emb x) (Emb y) = Emb (x <> y)
promoteOp _ _ = error "'promoteOp' not defined"
-- (retain the first case and define the other cases)

-- Task 1.2: Define the appropriate `Semigroup` and `Monoid`
-- instances for `Promote s`, using `promoteOp` as the monoid
-- operation.

instance (Semigroup s) => Semigroup (Promote s) where
  (<>) = error "'<>' for 'Promote s' not defined"

instance (Semigroup s) => Monoid (Promote s) where
  mempty = error "'mempty' for 'Promote s' not defined"
  mappend = error "'mappend' for 'Promote s' not defined"

-- Task 1.3: Write two predicates that QuickCheck can use
-- to test if a given monoid operation is lawful.
-- The first should test that the monoid operation is
-- associative, while the second should test that the identity
-- element satisfies the required laws.

testAssoc :: (Monoid g, Eq g) => g -> g -> g -> Bool
testAssoc x y z = error "'testAssoc' not defined"

testUnit :: (Monoid g, Eq g) => g -> Bool
testUnit x = error "'testUnit' not defined"

-- HINT: once you complete 1.3., you'll be able to invoke e.g.
-- `quickCheck (testUnit :: Promote String -> Bool)`
-- to check your work on the instances above.



-- PART 2 --

-- In this part you will show that any ordered type gives rise
-- to a monoid where the monoid operation corresponds to
-- taking the minimum.

data MinMonoid t
  = EmbMM t
  | Infinity
  deriving (Show, Eq)

instance (Arbitrary t) => Arbitrary (MinMonoid t) where
  arbitrary =
    (\f -> \x -> f x) <$> oneof [return (const Infinity), return EmbMM]
                      <*> arbitrary

-- Task 2.1: Define a monoid operation `minMonoidOp` (`<>`)
-- in such a way that given any list [x1,x2,...,xn] of values,
-- evaluating `EmbMM x1 <> EmbMM x2 <> ... <> EmbMM xn`
-- results in `EmbMM xi`, where xi = minimum of [x1,x2,...,xn].

minMonoidOp :: (Ord t) => MinMonoid t -> MinMonoid t -> MinMonoid t
minMonoidOp = error "'minMonoidOp' not defined"


-- Task 2.2: Define Semigroup and Monoid instaces for `MinMonoid`
-- using the `minMonoidOp` defined above.

instance (Ord t) => Semigroup (MinMonoid t) where
  (<>) = error "'<>' for 'MinMonoid' not defined"

instance (Ord t) => Monoid (MinMonoid t) where
  mempty = error "'mempty' for 'MinMonoid' not defined"
  mappend = error "'mappend' for 'MinMonoid' not defined"

-- HINT: A correct implementation will pass the test below.

testIsMinimum :: [Int] -> Bool
testIsMinimum xs
  | null xs = mconcat (map EmbMM xs) == Infinity
  | otherwise = mconcat (map EmbMM xs) == EmbMM (minimum xs)



-- PART 3 --

-- The merge operation takes two sorted lists as input,
-- and produces a single list as output, containing
-- all the elements of the input lists in sorted order.
-- Merge is famously used as a subroutine in various
-- sorting algorithms, including MergeSort.
-- E.g. merging the lists [1,3,6] and [2,5,6] would yield
-- the list [1,2,3,5,6,6].

-- The `SortList` type is intended to represent lists
-- that are sorted in _ascending_ order.

data SortList a = SortList [a] deriving (Show, Eq)

instance (Arbitrary a, Ord a) => Arbitrary (SortList a) where
  arbitrary = sortList <$> arbitrary

-- The `SortList` constructor is not supposed to be invoked
-- directly. Instead, we use the "smart constructor"
-- `sortList` defined below. This smart constructor ensures
-- that lists of type `SortList a` are indeed always sorted
-- in ascending order.

sortList :: (Ord a) => [a] -> SortList a
sortList xs = SortList (sort xs)

unSortList :: SortList a -> [a]
unSortList (SortList xs) = xs

-- Task 3.1: Define, as a predicate, what it means for a list to be sorted
-- in ascending order.
isSortList :: (Ord a) => [a] -> Bool
isSortList xs = error "'isSortList' not defined"

-- This is the invariant describing the SortList type:
-- a list obtained from any `SortList a` value should indeed
-- be a list sorted in ascending order.
sortedInv :: (Ord a) => SortList a -> Bool
sortedInv = isSortList . unSortList

-- Task 3.2: Define the merge operation `<|>` on SortList lists.

(<|>) :: (Ord a) => SortList a -> SortList a -> SortList a
(<|>) = error "'<|>' not defined"

-- Task 3.3: Write a predicate that tests whether the `<|>`
-- operation preserves the `isSortList` invariant defined above,
-- i.e. that merging two sorted lists results in a sorted list.

testSortList :: SortList Int -> SortList Int -> Bool
testSortList = error "'testSortList' not defined"

-- Task 3.4: Define a semigroup and monoid instance for `SortList a`
-- with monoid operation <|>. Write tests for the monoid laws.

instance (Ord t) => Semigroup (SortList t) where
  (<>) = error "'<>' for 'SortList t' not defined"

instance (Ord t) => Monoid (SortList t) where
  mempty = error "'mempty' for 'SortList t' not defined"
  mappend = error "'mappend' for 'SortList t' not defined"

testAssocSortList :: SortList Int -> SortList Int -> SortList Int -> Bool
testAssocSortList = error "'testAssocSortList' not defined"

testUnitSortList :: SortList Int -> Bool
testUnitSortList = error "'testUnitSortList' not defined"

